-- | "Chords" contains definitions of basic musical elements

{-# LANGUAGE GeneralizedNewtypeDeriving  #-}
{-# LANGUAGE DeriveDataTypeable  #-}
 

module Chords where
import Data.Generics
import System.Random
import Data.List hiding (transpose)
import System.Random


shuffle :: RandomGen g => g -> g
shuffle g = snd(next g)

swap :: (a->b->c) -> (b->a->c)
swap f a b = f b a

diffs :: (Num a, Ord a) => [a] -> [a]
diffs s = [y-x|x<-s, y<-s, y>x]


bracket :: a -> [a]
bracket x = [x]


-- | chromatic stuff first

c_period::Int
c_period = 12

newtype C_Note = C_N Int
    deriving (Typeable, Data, Show, Eq, Ord, Num, Enum)

instance Random C_Note where
    randomR (C_N a, C_N b) g = ( C_N (fst r), (snd r) )
        where r = (randomR (a, b) g)
    random g = ( C_N (fst r), (snd r) )
        where r = random g

newtype C_Int = C_I Int
    deriving (Typeable, Data, Show, Eq, Num, Ord)

c_Int :: C_Note -> C_Note -> C_Int
c_Int (C_N a) (C_N b) = C_I (b-a)

newtype C_ChordShape = C_CS [C_Int]
    deriving (Typeable, Data, Show)

newtype C_ChordRep = C_CR [C_Note]
    deriving (Typeable, Data, Show)

stripNoteRep :: C_Note -> Int
stripNoteRep (C_N n)=n

stripChordRep :: C_ChordRep -> [Int]
stripChordRep (C_CR c) = map stripNoteRep c

-- |a job for generic stuff methinks!
transposeNote :: C_Int-> C_Note  -> C_Note
transposeNote (C_I i)  (C_N n)= C_N (n+i)

transpose :: Data a => C_Int -> a -> a
transpose i cr = everywhere (mkT (transposeNote i)) cr

putChordShape :: C_ChordShape -> C_Note -> C_ChordRep
putChordShape (C_CS cs) n  = C_CR $ map (swap transpose n) cs

iSet :: C_Int -> [C_Int]
iSet = bracket

nSet :: C_Note -> [C_Note]
nSet = bracket

calcInts :: [C_Note] -> [C_Int]
calcInts ns  = [C_I (y-x) | (C_N x)<-ns, (C_N y)<-ns, y>x]

-- |need to sex this up to deal with chord-shape?
-- maybe also need to extract all C_Notes and take differences...
intervals :: Data a => a -> [C_Int]
intervals s =  (everything (++) (mkQ [] iSet) s) 
              ++
              (calcInts (everything (++) (mkQ [] nSet ) s))

chordIntervals :: Data a => a -> [C_Int]
chordIntervals = diffs . intervals

-- |harmony stuff

-- |a measure of consonance

strongConsonances :: [ C_Int ]
strongConsonances = [0,7,12,19,24]

weakConsonances :: [ C_Int ]
weakConsonances = [3,4,9,10,15,16,21,22]

consonantInts :: [C_Int]
consonantInts = [3,4,7,9,10]

newtype ConsVal = CV Int

data Consonance = C_Strong | C_Weak | C_Diss
                  deriving (Typeable)

consonance :: C_Int -> Consonance
consonance i = if elem i strongConsonances
               then C_Strong 
               else if elem i weakConsonances
               then C_Weak
               else C_Diss

cSet :: Consonance -> [Consonance]
cSet = bracket

-- | extracts plain consonances, and consonances derived from intervals
consonances :: Data a => a -> [Consonance]
consonances s = (everything (++) (mkQ [] cSet) s)
               ++
               (map (consonance) (intervals s))


-- | maybe this would be better as a vector? (store weak and diss separately)
newtype D_Val = D_V Int  
    deriving (Eq, Show, Num, Typeable, Data, Ord)

dissonance :: Consonance -> D_Val
dissonance C_Strong = 0;
dissonance C_Weak = 1;
dissonance C_Diss = 4;

-- | no need to genericize this I think
dissonanceTrace :: [Consonance] -> D_Val
dissonanceTrace cs = sum (map dissonance cs)

dSet :: D_Val -> [D_Val]
dSet = bracket

-- |extracts all dissonances from ints or direct diss values, or conses 
dissonances :: Data a => a -> [D_Val]
dissonances s = (everything (++) (mkQ [] dSet) s)
               ++
               (map dissonance $ consonances s)

d_score :: Data a => a -> D_Val
d_score s =  foldl (+) 0 (dissonances s)

-- |try to work out basic chromatic consonance stuff

newtype R_Val = R_V Int
    deriving (Show, Ord, Eq)

-- |maybe it would be better to have a codimensionally-defined unrelatedness function...
-- should probably deal with part-leading as well here at some point...
relatedness :: C_ChordRep -> C_ChordRep -> R_Val
relatedness (C_CR a) (C_CR b) = R_V 
                                (
                                 (length $ intersect a b) --number of notes in common
                                 + (length $ smartIntersect (intervals a) (intervals b))
                                )

newtype T_Val = T_V Int
    deriving (Show)

-- | will need to elaborate on this at some point
subordinate :: C_ChordRep -> C_ChordRep -> Int
subordinate f t = if d_score f < d_score t 
                  then 1
                  else 0

{-|
for ideal link, want x to be more related to its parent than any other in its arch
-}

-- | tries to take numbers into account
smartIntersect :: Eq a => [a] -> [a] -> [a]
smartIntersect l m = concat res
    where
      reps = nub $ intersect l m
      res = map (\x->replicate (min (count x l) (count x m)) x) reps

count :: Eq a => a -> [a] -> Int
count x s = length $ elemIndices x s



newtype C_PitchClass = C_PC Int

newtype C_Chord = C_C [C_PitchClass]
newtype D_Note = D_N Int
data Scale = Scale {tonic::C_Note,s_set::C_Note -> Bool}
newtype D_Int = D_I Int

classrep_above :: Int -> Int -> Int
classrep_above pc n = if (nc < pc) then
                          n + (pc - nc)
                      else
                          n + (c_period+pc-nc)
                          where
                            nc = n `mod` c_period

cycleRootUp :: C_ChordRep -> C_ChordRep 
cycleRootUp (C_CR (n@(C_N nint):ns)) = if n> (maximum ns)
                                       then C_CR $ ns ++ [n]
                                       else cycleRootUp (C_CR ((C_N (nint+c_period)):ns))

cycleRootDown :: C_ChordRep -> C_ChordRep 
cycleRootDown (C_CR k) = C_CR $ map negate $ reverse $ k2
                         where
                           (C_CR k2) = cycleRootUp $ C_CR $ reverse (map negate k)

recurse :: (a->a) -> Int -> a -> a
recurse f n x = (iterate f x)!!n

randomInversion :: RandomGen a => C_ChordRep -> a-> C_ChordRep
randomInversion c@(C_CR ints) g = if (up==0)
                                  then recurse cycleRootUp n c
                                  else recurse cycleRootDown n c
    where 
      l = length ints
      (n ,g1) = randomR (1,length ints) g
      (up,g2) = randomR (0::Int,1) g1

--randomInversion (C_CR []) g = (C_CR [])
--randomInversion c@(C_CR ((C_N arep):as)) g = C_CR $ (C_N (((arep+rnd*12) `mod` 48)-24)):asnew
--    where
--      (rnd,g1)=randomR (0::Int,3) g
--      (C_CR asnew) = randomInversion (C_CR as) g1 

-- | four-note chords for now
randomChord :: RandomGen a => a -> C_ChordRep
--randomChord g = C_CR $ sort $ fst $ randomFromList ([], range) 4 g 
--    where
--      range = [-12..12]


--more consonant random chord

randomChord g = C_CR [c1,c2,c3,c4]
    where 
      (n1,g1)= randomR (-12,12) g
      (n2,g2)= ( randomR (0,4) g1 )
      (n3,g3)= ( randomR (0,4) g2 )
      (n4,g4)= ( randomR (0,4) g3 )
      c1=n1
      c2=transpose  (weakConsonances!!n2) c1
      c3=transpose  (weakConsonances!!n3) c2
      c4=transpose  (weakConsonances!!n4) c3


{-
--random chord from given set of shapes
admissableShapes :: [C_ChordShape]
admissableShapes = [majorShape,minorShape]

randomChord g = putChordShape (admissableShapes!!n2) n1
    where
      (n1,g1)= randomR (-12,12) g
      (n2,g2)= ( randomR (0,(length admissableShapes)-1) g1 )
--could start with a random inversion, but whatever
-}

randomFromList :: (RandomGen r, Eq a) => ([a], [a]) -> Int -> r -> ([a],[a])
randomFromList l 0 _ = l
randomFromList (as,bs) n g = randomFromList (newas,newbs) (n-1) newg
    where
      l=length bs
      (r,newg)=randomR (0,(length bs)-1) g
      newas = (bs!!r):as
      newbs = delete (bs!!r) bs
                            

rotate :: [a] -> [a]
rotate [x]=[x]
rotate (x:ys) = ys ++ [x]

majorShape :: C_ChordShape
majorShape = C_CS [4,3]

minorShape :: C_ChordShape
minorShape = C_CS [3,4]

-- | some actual data

major :: C_ChordShape
major = C_CS [0,4,7]

minor  :: C_ChordShape
minor = C_CS [0,3,7]